#ifndef CLASS_ALIGN5_HH
#define CLASS_ALIGN5_HH

// header file for the ALIGN5 class library
// =========================================================
// Author: Martin Braeuer 
//
// NOTE: The name is derrived from the alignment methods:
//       ALIGN 1 => Target-spot pre-tracking alignment
//       ALIGN 2 => pre tracking alignment using seed/ 
//                  confirmation combis
//       ALIGN 3 => After tracking alignment, using 
//                  accepted hits, having seed/confirmation 
//                  combis==ResidualElements==REle's, 
//                  lateral alignment
//       ALIGN 4 => As No. 3, having rotational and longi-
//                  tudinal degrees of freedom
//       ALIGN 5 => Storing of the REle's in matrices
//
// Modification history:
// MBr  27.05.99: Porting from align5.cc V2.0 to class
// MBr  14.09.99: Porting class to VDSLIB 2-19 
//                    => align5c1-0
// MBr  30.10.99: unsused vars removed, acess to matrix 
//                and vector, better printing
//                    => align5c1-1
// MBr  31.10.99: fill the matrix with user routines
//                and vector, better printing
//                    => align5c1-2
// MBr  17.11.99: improved par-order, bad-chip-side mechanism
//                    =>
// MBr  30.11.99: bad-chip-side mechanism to other Class
//                    => align5c1-3
// =========================================================


#include "Struct_Gede.h"   // use Gede_t structs
#include "Struct_Seed.h"   // use Seed_t structs
#include "Struct_Hitb.h"   // use Hitb_t structs
#include "Struct_Track.h"  // use Track_t structs

#include "Class_GEDE.h"    // geometry information
#include "Class_SEED.h"    // seed generation
#include "Class_HOLMES.h"  // holmes track finding


// prototypes for LAPACK:
extern "C" {
  int dsytrd_(char *, int *, double *, int *, double *, double *, double *,
	      double *, int *, int *);
  int dsterf_(int *, double *, double *, int *);
  int dgelsx_(int *, int *, int *, double *, int *, double *, int *, int *, 
	      double *, int *, double *, int *);
  int dgelss_(int *, int *, int *, double *, int *, double *, int *, double *,
	      double *, int *, double *, int *, int *);
//  dinv_(int *, double *, int *, double *, int *); // ALready in HOLMES..
}

// public definitions:
// constant values for dofit :

#ifndef Pi
 #define Pi 3.14159265    // An unknown Constant..
#endif

#define LATERAL    1
#define ROTATIONAL 2
#define AXIAL      4

// common consts :
#define LENSC     10000             // from cm to my
#define NSEED      4                // number of seed-strips
#define MAXPLANE  24                // max No of SL's per track
#define MAXGEDE  160                // max no of GEDE entries, with 10 SL's
#define NRES     (MAXPLANE - NSEED) // max no of Residual planes
#define NALIGN   NSEED+1            // no of alig pars/res-combi (5 planes)
#define LAY_N1 1                    // layer ID of first  n-side 
#define LAY_P1 2                    // layer ID of first  p-side 
#define LAY_P2 3                    // layer ID of second p-side 
#define LAY_N2 4                    // layer ID of second n-side

// the minimal number of gathered events:
#define MINGATHER 15

// define storage numbers:
#define ALIGN5_MAXP  255
#define ALIGN5_MAXS  265
#define ALIGN5_RPO   (ALIGN5_MAXP+1)
#define ALIGN5_CPO   (ALIGN5_MAXP+2)
#define ALIGN5_TPO   (ALIGN5_MAXP+3)
#define ALIGN5_SMAT  1
#define ALIGN5_SVEC  2

// the class definition for align5 -------------------------------------------

class ALIGN5_class {
private:
  // functions
             
       // Print an error message to the user:
       void fError_out(const char *);

       // invert a n*n matrix
       int sqrinvd(double *a, int n);                          

       // calculate the eigenvalues
       int eigenvalues(double *matrix, double *result, int n);

       // complete orthogonal decomp.
       int solveortho(double *, double *, double *, int);  

       // sing. val decomp.
       int solvesvd(double *, double *, int, double *, double, int *, int *    );

  // types :
  // remaining pars per plane :
       typedef struct{
	 int    rc;       // no of reles related to plane
	 double rsum;     // sum of residuals for this plane
	 double paru;     // LATERAL ALIGNMENT PARAMETER
	 double pars;     // LATERAL ALIGNMENT PARAMETER
	 double parz;     // LATERAL ALIGNMENT PARAMETER
	 double sparu;    // LATERAL ALIGNMENT PARAMETER
	 double spars;    // LATERAL ALIGNMENT PARAMETER
	 double sparz;    // LATERAL ALIGNMENT PARAMETER
	 int    matPu;    // number of parameter element for du
	 int    matPs;    // number of parameter element for dalpha
	 int    matPz;    // number of parameter element for dz
	 int    status;   // status same as gede.status, but gedeadd can be indiced
       } GedeAdd_t;

       // map for seedings : (Seed_mi from Class_SEED.h)
       typedef multimap< float, Seed_t*, less<float> >::iterator Seed_mi;

       // hitb map for each track : (hits will be checked across many seedings!
       typedef multimap< int, Hitb_t*, less<float> >::iterator thitb_mi;

  // data
       HOLMES_class *HOLMESp;         // Access pointer to HOLMES  
       SEED_class   *SEEDp;           // Access pointer to SEED
       GEDE_class   *GEDEp;           // Access pointer to GEDE
       GedeAdd_t    gedeadd[MAXGEDE]; // Fast access to additional 
                                      // plane related stuff

       int    prepare;      // 1 if prepare step was performed, 0 else
       int    gather;       // 1 if gather  step was performed, 0 else
       int    solve;        // 1 if solve   step was performed, 0 else
       double **pars;       // pointer to parameter as a property of gede
       double *avec;        // pointer to the vector of sums
       double *svec;        // pointer to the vector of corrections
       double *esvec;       // pointer to the vector of corrections errors
       double *svvec;       // pointer to the eigen values of amat    
       double *amat;        // pointer to the matrix of sums
       double *atam;        // pointer to the inverse matrix of sums
       double *avsing;      // pointer to the vector of sing values (increasing order)
       int    rank;         // effective rank of problem, rank<= npars
       int    *mp;          // vector with col1,row indizes of amat
       int    *hiscount;    // pointer to the vector of hitsums per par
       Gede_t **gede;       // gede gede of this par
       int    *type;        // type of this par
       int    dofit;        // 1: ufit, 2:s-fit,4:z-fit
       int    holmesquad[4];// quadrant parameter for holmes
       double sum0;         // initial Residual-sum (chi^2)
       int    dgf;          // number of degrees of freedom
       
       int    npars;        // number of parameter to align
       int    ntrack;       // number of used tracks
       int    nreles;       // number of used reles
       double fichi2;       // final   chi^2
       
      /* store retrieve: book, row, column, value
          book = ALIGN5_SMAT  ALIGN5_SVEC          
      */
      int (*Pstore)(   int, int, int, double  ); // pointer to store function ==0 -> int. store
      int (*Pretrieve)(int, int, int, double *); // pointer to retr. funtion  ==0 -> int. storage 

public:
  // constructors
       ALIGN5_class(); 
  // constructor with parameters:
       ALIGN5_class(HOLMES_class *HOLMESp,
                    int ,int ,int ,int *, 
		    int (*)(int, int, int, double  ),
		    int (*)(int, int, int, double *)  ); 
					 
  // Functions
       /* ==================================================================
        * == fClear : Clear/unbook all allocated structures               ==
        * ==                                                              ==
        * ================================================================== 
        */
        void fClear();


       /* ==================================================================
	* == fGetPars : Get number of parameters                          ==
	* == needs : prepare has to run before (0 value otherwise)        ==
	* ================================================================== 
	*/
       int fGetPars();

       /* ==================================================================
	* == fGetPars : Get the pars x pars Matrix                        ==
	* == needs : gather has to run before (0 value otherwise)         ==
	* ================================================================== 
	*/
       double *fGetMatrix();

       /* ==================================================================
	* == fGetMatrix : Get the pars size Vector                        ==
	* == needs : Gather has to run before (0 value otherwise)         ==
	* ================================================================== 
	*/
       double *fGetVector();

       /* ==================================================================
	* == fGetChanges : Get the pars size Vector with solution         ==
	* == needs : Solve has to run before (0 value otherwise)          ==
	* ================================================================== 
	*/
       double *ALIGN5_class::fGetChanges();
	 
        /* ==================================================================
	* == fGetTracks : Get number of processed Tracks                  ==
	* == needs : Gather has to run before (0 value otherwise)         ==
	* ================================================================== 
	*/
       int fGetTracks();

       /* ==================================================================
	* == fGetReles : Get number of Residual elements                  ==
	* == needs : Gather has to run before (0 value otherwise)         ==
	* ================================================================== 
	*/
       int fGetReles();

       /* ==================================================================
	* == fGetDgf : Get number of degrees of freedom                   ==
	* == needs : Gather has to run before (0 value otherwise)         ==
	* ================================================================== 
	*/
       int fGetDgf();

       /* ==================================================================
	* == fGetInichisq : Get initial chi^2                             ==
	* == needs : Gather has to run before (0 value otherwise)         ==
	* ================================================================== 
	*/
       double fGetInichisq();

       /* ==================================================================
	* == fGetFinchisq : Get initial chi^2                             ==
	* == needs : Solve has to run before (0 value otherwise)          ==
	* ================================================================== 
	*/
       double fGetFinchisq();

       /* ==================================================================
	* == fGetParmap : Get text map of parameter list                  ==
	* == needs : Prepare has to run before (0 value otherwise)        ==
	* ================================================================== 
	*/
       int fGetParmap();

       /* ==================================================================
        * == fPrepare : Prepare structures for align5                     ==
        * == Parameter: Pointer to HOLMES structure                       ==
        * ==            Lateral-flag (0=Do not align lateral)             ==
        * ==            Axial-flag (0=Do not align in z)                  ==
        * ==            Rotational-flag (0=Do not align stereo angle)     ==
        * ==            Array with plane ID's wich stay fixed             ==
        * ==                                                              ==
        * == needs : ----                                                 ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fPrepare(HOLMES_class * ,    // pointer to HOLMES
		    int ,               // align lateral
		    int ,               // align Axial
		    int ,               // align Rotational
		    int *,              // array of fixed sides
		    int (*store)(   int, int, int, double  ),
		    int (*retrieve)(int, int, int, double *) ); // store pointer

       /* ==================================================================
        * == fGather : Accumulate the infos from tracks                   ==
        * == needs   : fPrepare has to be run before                      ==
        * ==           int tr : track no in HOLMES current list           ==
        * ==                                                              ==
        * ================================================================== 
        */
       int fGather(int tr);      // build matrix elements from projectors

       /* ==================================================================
        * == fSolve : Invert the algnment matrix, calc the pars           ==
        * == needs : fGather has to be run before at least 10 times       ==
        * ==                                                              ==
        * == Using SVD to get the result!                                 ==
        * ================================================================== 
        */
       int fSolve();             // calculate parameters

       /* ==================================================================
	* == fChange_GEDE : Change the entrire GEDE structure with        ==
	* ==                alignment corrections                         ==
	* == needs : fSolve has to be run before                          ==
	* ==                                                              ==
	* ================================================================== 
	*/
       int fChange_GEDE();


       /* ==================================================================
        * == fOut   : Output of the alignment parameters                  ==
        * == needs : fSolve has to be run before                          ==
        * ==                                                              ==
        * ==                                                              ==
        * ================================================================== 
        */
       int   fOut();      // output gede ff

       /* ==================================================================
        * == fCalcchisq : Calculate a chi^2 from the data for new pars    ==
        * == needs : gather has to be run before                          ==
        * == return: The actual chi^2 value                               ==
        * ==                                                              ==
        * ================================================================== 
        */
       double fCalcchisq();  // calc a new chisq

       /* ==================================================================
        * == fUnsin :  Calculate the 7bit ID of sides:                    ==
        * == needs  : --                                                  ==
        * == pars   : sl=superlayer, sec=sector, lay=layer (1..count)     ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fUnsin(int sl,int sec,int lay) { 
	 int i = (lay-1)+(4*(sec-1))+(16*(sl-1));
	 return(i); 
       }

       /* ==================================================================
        * == fSideUnsin :  Calculate the 7bit ID of side                  ==
        * == needs  : --                                                  ==
        * == pars   : side                                                ==
        * == return : The USID (0..127)                                   ==
        * ==                                                              ==
        * ================================================================== 
        */
       inline int fSideUnsin(int side) {
	 if (side < 20000)
	   return((side%10) + 4*((side%1000)/300) + 16*(  int(side/10)%10)); 
	 else 
	   return((side%10) + 4*((side%1000)/300) + 16*(1+int(side/10000))); 
       }


  // Destructor:
       ~ALIGN5_class();

  // data public members
    // NONE there!
}; // ============== END of Class_ALIGN5 def. ! ==============================

#endif




















