//                              -*- Mode: C++ -*-
//: EMCellFraction.hh
//
//
// Author: Mauro Villa   
// ECAL group 
// HERA-B experiment, Desy
// e-mail: villa@bo.infn.it
//
// Created 3/4/97
// Last Update ........
//

#ifndef EMCELLFRACTION_HH_
#define EMCELLFRACTION_HH_

#include "care/EMHitCell.hh"

/**
class EMCellFraction.
  Concrete derived class for the hierarchical cluster tree used in
  pattern recognition of ECAL. This class is a "leaf": it does not have
  children, it cannot receive Add, Move or Remove message.
  It is used to store a part of a cell energy.
*/

class EMCellFraction : public EMHitCell {
public:
  /**@name Constructors and Destructor */
  //@{
  /** Constructors.
     The construction of EMCellFraction is done referring to some
     parent. The depth level is adjusted accordingly. */
  EMCellFraction( EMHitCell* hitc, Energy_t fraction );
  ///  The default constructor should be used only by the HCBuilder class.
  EMCellFraction();
  /// Destructor: a virtual destructor is needed
  virtual ~EMCellFraction();

  //@}

  /**@name Status access functions */
  //@{
  //--
  /// Return the maximum Energy of the Cell itself. 
  virtual Energy_t  maxEnergy() const;

  //--
  /// Access the energy of a hit cell
  virtual Energy_t  energy() const;

  //--
  /// Access the squared energy error of a hit cell
  virtual Energy_t  sqerror() const;

  //--
  /// Get the fraction energy
  inline Energy_t getFraction() const;
  //
  /// Get the linked EMHitCell
  inline EMHitCell* getHitCell() const;

  //
  //--
  ///  Dump for debugging.
  virtual void dump() const;
  //@}

  //--
  /** Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
   that are object with definite analyzing properties. 
   A direct call to the doAnalysis() member of EAClusterAnalyzer for this
   "leaf" is done. */
  virtual void analyzeMe(EAClusterAnalyzer & analyzer) const;

  //--
  /** Analysis and traversing of the cluster tree is done by EAClusterAnalyzers
   that are object with definite analyzing properties. 
   A direct call to the doModify() member of EDClusterModifier for this
   "leaf" is done. */
  virtual void modifyMe(EDClusterModifier & modifier);

  //--
  /// Set the internal values of the cell fraction
  inline EMCellFraction* pointToCell(EMHitCell* cell, Energy_t fract);

protected:

private:
  /**@name Private Data */
  //@{
  ///  This is a link to the HitCell entry 
  EMHitCell*  hitc;
  ///  This is the fraction of energy stored in this cell
  Energy_t fraction;
  //@}
};


#include "care/EMCellFraction.icc"
#endif // EMCELLFRACTION_HH_




