/************************************************************************
* This routine translates GEANT output hits into ARTE tables
* HITB and DIGB for OTR modular geometry.
*
* Called from DGEVNT.
*
* Written by S.Nowak, C.Stegmann, A.Lanyov     Oct 1999.
*
* Updated: hitb->chid is filled       A.Lanyov Dec 1999.
*
* Arguments  kind         description
*-----------------------------------------------------------------------
* IPL        in/out       Detector number in GEDE bank
************************************************************************/
#include <vector.h>
#include <math.h>
#include "arte/arte.hh"
#include "arte/ArteMessage.hh"
#include "gean/ZGESL.h"
#include "digi/zgdig.hh"
#include "digi/dglev.h"
#include "digi/dgnoise.h"

extern "C" {
  void dgomod_(int &);
  void grndm_(float [], const int &);
  void granor_(float &, float &);
  void gpoiss_(float *,int * ,int &);
  void sortzv_(int[], int[], int&, int&, int&, int&);
  int otr_gede_status(int, int, int);
}

float otr_pathlen(const float, const float,const float); 
float first_cluster_position(const float);
void  otr_crosstalk(const int,const int,const ArtePointer<GEDE>,vector<digx> *,
      const float);
void  otr_noise(const ArtePointer<GEDE>,vector<digx> *);


void dgomod_(int & ipl) {
   int i0=0, i1=1, mi1=-1; 
   float ydg10 = 2.5, rndm[3];
   vector<digx> *dg_otr[4];
   for(int i=0; i<4; i++)    dg_otr[i] = NULL;

/* In GEAN96 all MIMP's of a module are assigned to the first GEDE of this 
   module. Therefore one has to calculate the number of GEDE's belonging 
   to this module (nsec) */
   ArtePointer<GEDE> gd(ipl);
   unsigned int j=0;
   for(j=ipl+1;;j++) {
       if(j > ArteTable<GEDE>::size()) break; 
       if(ArtePointer<GEDE>(j)->gcop != gd->gcop ) break;  
   }
   int nsec = j - ipl;
   if(nsec > 4) { 
     arteMsg << "dgomod: in GEDE plane "<<ipl<<" there are " <<nsec<<" sectors";
     arteMsg.error();
   }

// define some often used quantitites identical for all sublayers
   zgdig ZGDIG(ipl);
   float rout =  gd->pit/3.;              // cell''s outer radius
   float rin  =  0.8660254*rout;          // cell''s inner radius

// calculate number of mimps 
   int nmim=0, ifmimp=0;
   nmim = ZGDIG.Get_nmimp(ifmimp);

// prepare intermediate storage
   digx otrdg;
   for(int i=0; i<nsec; i++)    dg_otr[i] = new vector<digx>;

/*The x coordinate will be
       transformed into the plane zmin of the given detector and rotated */

// loop over all MIMPs
   for(int ih=0;ih<nmim;ih++) {
      ArtePointer<MIMP> mp(ih+ifmimp);
      float x0[3],  x_rot[3], tx0[3],  tx_rot[3]; 
      x0[0] = mp->x;
      x0[1] = mp->y;
      x0[2] = 0.;  
      ZGDIG.Rotate2plane_Z(x0,x_rot);    //rotate x into wire orientation  
      tx0[0] = mp->tx;   
      tx0[1] = mp->ty;   
      tx0[2] = 0.; 
      ZGDIG.Rotate2plane_Z(tx0,tx_rot); //rotate tx into wire orientation
      float cx_rot = 0.;
      if(fabs(tx_rot[0]) < 1.e+10) cx_rot = 1./sqrt(1.+ tx_rot[0]*tx_rot[0]);

      float ytv = x_rot[1] + tx_rot[1]*(mp->z-gd->zmin); 
      float yth = x_rot[1] + tx_rot[1]*(gd->zmax-mp->z);

//    Loop over all sectors - check into which sector the mimp goes
      for(int j=0; j<nsec; j++) {
         ArtePointer<GEDE> gf(ipl+j);
         zgdig ZGDIG(ipl+j);

         if ( (ytv <= gf->symn || ytv >= gf->symx)    &&
              (yth <= gf->symn || yth >= gf->symx))   continue;   
         
         for(int itube=0;itube<gf->nsl;itube++){ // loop over monolayers =tubes 
	   if(gf->nwi[itube] <= 0 ) continue;
            float zwir = gf->zfw[itube];
            float y1w = gf->y1w[itube] + 0.5 * ydg10;
            float y2w = gf->y2w[itube] - 0.5 * ydg10;
            float yttv = (zwir - mp->z - rin) * tx_rot[1] + x_rot[1];
            float ytth = (zwir - mp->z + rin) * tx_rot[1] + x_rot[1];
            if (yttv < y1w && ytth < y1w)  break;
            if (yttv > y2w && ytth > y2w)  break;

            float xfw = gf->xfw[itube];
            int n_wire = gf->nwi[itube];
            float xtref = (zwir- mp->z) * tx_rot[0] + x_rot[0];
            float xttv = xtref - rin*tx_rot[0];
            float xtth = xtref + rin*tx_rot[0];

//          first/last cell hidden by the track 
            int nv = int((xttv-(xfw-rout))/gf->pit) + 1;
            int nh = int((xtth-(xfw-rout))/gf->pit) + 1;
            if(nv <= 0 && nh <= 0)  continue; 
            if(nv > n_wire && nh > n_wire) continue; 
            int isn = 0; 
            if(nv > nh) {
               isn=-1;
               if(nv > n_wire) nv = n_wire;
               if(nh < 1) nh = 1;
            }
            else {
               isn = 1;
               if(nv < 1) nv = 1;
               if(nh > n_wire) nh = n_wire;
            }

//          loop on wires
            for (int n=nv;n!=nh+isn;n=n+isn) {          

	      // Check status of wire and ignore the hit if this wire is bad.
	      if(dglev_.idglev[LGESL_WCH-1] == 3) {  // dglev OTR 3
		int status = otr_gede_status(ipl+j, itube, n-1);
		if (status) continue;
	      }
	      
               float xwir = xfw + (float(n)-1.)*gf->pit;
               float dd = (xtref-xwir) * cx_rot;   // smallest distance to wires
               grndm_(rndm,i1);
               if(rndm[0] > gf->eff) continue;
//             calculate path length in cell 
               float dpath = otr_pathlen(xttv-xwir,xtth-xwir,rin);
               if (dpath <= 0.) continue;    //  track does not hit cell
               if(dglev_.idglev[LGESL_WCH-1] != 0) {    //long version
                  float dlen = first_cluster_position(dd);       
//                if length needed for the first cluster is longer then half
//                the path length in the cell then no hit is produced
	          if (dlen > 0.5*dpath) continue;
//                new  distance to wire 
                  float ddion = sqrt( dd*dd + dlen*dlen); 
//                simulate electronic and mechanical smearing
                  granor_(rndm[0],rndm[1]);
                  ddion =  ddion + gd->acc*rndm[0];
                  if(fabs(ddion) > rout) continue;
                  if(fabs(dd) > gf->dod) continue;
                  int knum = itube << 16;
                  otrdg.strip = knum + n;                    // strip number
                  otrdg.charge= ddion;                       // distance
                  otrdg.mimp = mp.index();                   // mimp number
                  ArtePointer<MTRA> mt = mp->mtra;
                  otrdg.track = mt.index() ;                 // track number
                  dg_otr[j]->push_back(otrdg);

//                simulate cross talk
//                  otr_crosstalk(itube, n, gf, dg_otr[j], ddion);
	       }
               else  {                                //short version
                  granor_(rndm[0],rndm[1]);
                  dd = dd + gd->acc*rndm[0]; 
                  if (fabs(dd) > rout) continue;
                  if(fabs(dd) > gf->dod) continue; 
                  int knum = itube << 16;
                  otrdg.strip = knum + n;                    // strip number
                  otrdg.charge= dd + gd->acc*rndm[0];        // distance
                  otrdg.mimp = mp.index();                   // mimp number
                  ArtePointer<MTRA> mt = mp->mtra;
                  otrdg.track = mt.index() ;                 // track number
                  dg_otr[j]->push_back(otrdg);
	       }
  
            }       //======================== end of loop on wires
         }          //======================== end of loop on monolayers
	  
      }             //======================== end of loop on sectors 
   }                //======================== end of loop on mimps

// simulate noise
   if(dglev_.idglev[LGESL_WCH-1] != 0) {    //long version
      for (int j = 0; j< nsec; j++) {
          ArtePointer<GEDE> gf(ipl+j);
          otr_noise(gf, dg_otr[j]);
      }
   }

// preparation for sorting 

   int *dg_srt = NULL; int *ind = NULL; int *ind1 = 0;
   for (int j=0;j< nsec;j++) {  //======================== loop on sectors
      int ndigde = dg_otr[j]->size();
      if (ndigde == 0) {
         zgdig ZGDIG(ipl+j);
         ZGDIG.Set_hit_for_nomimp(); 
         continue;
      }
      ArtePointer<GEDE> gf(ipl+j);
      int max_wire = gf->nwi[0];
      for (int i = 1; i<gf->nsl; i++) {
         if(gf->nwi[i] > max_wire) max_wire = gf->nwi[i];
      }

      dg_srt = new int[ndigde];
      ind    = new int[ndigde];
      ind1   = new int[gf->nsl*max_wire];
      for(int i=0; i<ndigde; i++) {
         vector<digx>::iterator dg = dg_otr[j]->begin() + i;
         dg_srt[i] = dg->strip;
      }
      for (int i=0; i<ndigde; i++) ind[i]  = 0; 
      for (int i=0; i<gf->nsl*max_wire; i++) ind1[i] = 0;

//    sort the  matrix digx  due to increasing tube/wire number
      sortzv_(dg_srt,ind,ndigde,mi1,i0,i0);

//    set digit and hit bank
     int iflag = 0,   iflag1 = 0;
     for(int ih = 0; ih < ndigde; ih++) {

//      first the digits
        vector<digx>::iterator dg = dg_otr[j]->begin() + ind[ih] - 1;
        mmbook_(ArtePointer<DIGB>::getLink().getRef(),1);
        int nn1 =ArteTable<DIGB>::size() ;
        ArtePointer<DIGB> dgb(nn1);
        if(iflag == 0) {        // first digit in this detector
           gf->dig1 = nn1;        iflag = 1;           
        }
        gf->digl = nn1;

        int itube = dg->strip >> 16;
        int iwire = dg->strip&65535;
        dgb->id = 10000*((iwire-1)*4 + itube+1) + ipl + j;
//      id = 111 0068 means strip=28, tube=3, ipl+j=68
        dgb->q  = dg->charge; 
        if (dg->mimp != 0) 
           createRelation(ArtePointer<MIMP>(dg->mimp),dgb,&MIMP::digb);

//      next the hitb bank
        ArtePointer<HITB> htb;

        if(!ind1[itube*max_wire+iwire-1] ) {    //****book a new hitb bank
           mmbook_(ArtePointer<HITB>::getLink().getRef(),1);
           int nn = ArteTable<HITB>::size();
           htb = ArtePointer<HITB>(nn);
           if(!iflag1) {
             gf->hit1 = nn;    iflag1 =1;
           }
           gf->hitl = nn;
           htb->gede = ipl + j;
           htb->x  = gf->xfw[itube] + float(iwire-1) * gf->pit;
           htb->z  = gf->zfw[itube];
           htb->q  = dg->charge;           
           htb->r  = gf->acc;
           ind1[itube*max_wire+iwire-1] = nn;
           htb->chid = 32*itube + (iwire-1);
        }
           
        else {
           htb = ArtePointer<HITB>(ind1[itube*max_wire+iwire-1]);
           if(fabs(dg->charge) < fabs(htb->q)) htb->q = dg->charge;
           htb->chid = 32*itube + (iwire-1);
        }    

        if(dg->mimp) 
           createRelation (ArtePointer<MIMP>(dg->mimp),htb,&MIMP::hitb);
	createRelation (htb,dgb,&HITB::digb);
     }         //======================== end of loop on digitized hits

     if(dg_srt != NULL) { delete dg_srt;  dg_srt = NULL;} 
     if(ind    != NULL) { delete ind;     ind = NULL;} 
     if(ind1   != NULL) { delete ind1;    ind1= NULL;}

   }         //======================== end of loop on sectors

   for(int i=0; i<4; i++) { 
       if(dg_otr[i] != NULL) { delete dg_otr[i];  dg_otr[i] = NULL;}
   } 
   ipl += nsec -1;
}

float otr_pathlen(const float xv,const float xh,const float rin) {
//                calculate path length in cell

   float xent[2], zent[2];
   int ix = 0;  // counter for crossings with cell
   float rout = 1.1547 * rin;              // 2/sqrt3
   float tx = (xh-xv)/(2.*rin);  
   float xlu =  - 0.5*rout;     // corners of a cell
   float xru =  + 0.5*rout;

   if (xv >= xlu && xv <= xru) {      // lower boundary
      xent[ix] = xv;
      zent[ix] = 0;
      ix++;
   }
   if (xh > xlu && xh < xru) {         // upper boundary
      xent[ix] = xh;
      zent[ix] = 2.*rin;
      ix++;
   } 
//   if(ix == 2} {}

// lower left side and upper right side ( same slope)
   float tcn = - 0.57735;
   float zc = 0.;
   if (tx != tcn) zc = (xlu - xv) / (tx - tcn);
   if (zc < rin && zc > 0.) {
      xent[ix] = tcn*zc + xlu;
      zent[ix] = zc;
      ix++;
   }
   zc = 0.;
   if (tx != tcn) zc = (xru - xh) / (tx - tcn) + 2.* rin;
   if (zc > rin && zc < 2.*rin) {
      xent[ix] = tcn*(zc-2.*rin) + xru;
      zent[ix] = zc;
      ix++;
   }

// lower right side and upper left side ( same slope)
   tcn = -tcn;
   zc = 0.;
   if (tx != tcn) zc = (xlu - xh) / (tx - tcn) + 2.*rin;
   if (zc > rin && zc < 2.*rin) {
      xent[ix] = tcn*(zc-2.*rin) + xlu;
      zent[ix] = zc;
      ix = ix + 1;
   }
   zc = 0.;
   if (tx != tcn) zc = (xru - xv) / (tx - tcn);
   if (zc < rin && zc > 0) {
     xent[ix] = tcn*zc + xru;
     zent[ix] = zc;
     ix++;
   }

   if (ix != 2) return 0.;    //  track does not hit cell

// calculate path length in cell
   return sqrt((xent[0]-xent[1])*(xent[0]-xent[1]) + (zent[0]-zent[1])*(zent[0]-zent[1]));

}

float first_cluster_position(const float dd) {
/* throw position of first cluster. the number of clusters per unit
   length depends on the distance from the wire ( attachment in drift space) */
  int i2 = 2;
//alam , blam :  number of clusters per unit length
  float alam =-58., blam =20., dlen = 0., rndm[2]; 
  float lam = alam*dd + blam;
  if (dd > 0.12) lam = 13.;
//  maximum of gamma distribution
  float xran=1., xad = 0.;
  while(xran > xad) {  
    grndm_(rndm,i2);
    dlen = rndm[0];
    xad = 2*lam*exp(-2.*lam*dlen);
     xran = 2.*lam*rndm[1];
  }
  return dlen;
}   


void otr_crosstalk(const int itube,const int nw,const ArtePointer<GEDE> gf,vector<digx> *dg_otr, const float ddion) {
// simulate cross talk
/*  neighboring sector ???
    itube   0 1 2 3
    sectors 1 2 3 4
    itnei   1 0 0 0                  */    

    int itnei, nnl, nnr; int i1=1;
    float fcross = 0.02; float rndm[2]; 
    digx otrdg;
    if (itube == 0) { 
       itnei = 1;
       nnl = nw - 1;
       nnr = nw;
    }
    else {
       itnei = 0;
       nnl = nw;
       nnr = nw + 1;
    }
  
//  cross talk. save the same measured time
    for (int in=nnl;in<=nnr;in++) {
       if (in < 0 || in > gf->nwi[itnei-1]) continue;
       grndm_(rndm,i1);
       if (rndm[0] < fcross) {
//        smear cross talk hit by electronics noise
          granor_(rndm[0],rndm[1]);
          int knum = itnei << 16;
          otrdg.strip = knum + in      ;             // strip number
          otrdg.charge = ddion + rndm[0]*gf->acc;    // distance
          otrdg.mimp = 0;                            // mimp number
          otrdg.track = 0;          ;                // track number
          dg_otr->push_back(otrdg);
       }
    }
}

void otr_noise(const ArtePointer<GEDE> gf,vector<digx> *dg_otr) {
   int noise[1]={0}; int i1=1, i2=2;
   float noise_mean[1], rndm[2];
   digx otrdg;

   for(int itube = 0; itube < gf->nsl; itube++) {
      noise_mean[0] =gf->nwi[itube]*dgnoise_.noise[gf->cmp-1]; 
      if(noise_mean[0] < 0.)      noise[0] = int(-noise_mean[0]); 
      else if(noise_mean[0] > 0.) gpoiss_(noise_mean,noise,i1);
      int knum = itube << 16;
//    number of hits
      for (int ih=0; ih < noise[0]; ih++) {
         grndm_(rndm,i2);
         int nw = int(float(gf->nwi[itube])*rndm[0]) +1;
         if (nw > gf->nwi[itube]) nw = gf->nwi[itube];
         otrdg.strip = knum + nw;                    // strip number
         otrdg.charge= gf->pit*rndm[1]/3.;           // distance
         otrdg.mimp = 0;                             // mimp number
         otrdg.track = 0 ;                           // track number
         dg_otr->push_back(otrdg);
      } 
   }
}
