// See high_pt.hh for details.

// Author: Vladislav Balagura (balagura@hera-b.desy.de, 16.05.2000)

#include "ptset/high_pt.hh"
#include "ptset/global_error.hh"
#include "ptset/setup_tools.hh"
#include "ptset/setup_menu.hh"

HPT_Setup hpt_setup;

/* The following is needed to make interactive command /create
   which reads the chamber names in the following format:
     /create{ {{i1- i1+}{o1- o1+}}
              {{i2- i2+}{o2- o2+}}
	      {{i3- i3+}{o3- o3+}} }
   This means: first superlayer {{i1- i1+}{o1- o1+}} is composed of 2 parts,
   part one (inner high-pt) has chambers i1-, i1+; part two (outer) - o1-, o1+;
   and so on for the 2d and 3d superlayers.
   */   
enum Result {NEXT, END, ERROR};
Result get_block(istream& is, string* s) { // reads block { ... }
  // taking into account the presence of nested blocks: { .. {..} .. }
  char c;
  if ( !(is >> c) ) return END;
  if ( c != '{' ) {
    global_error()("can not read \'{\' in front of chamber names"); 
    return ERROR;
  }
  int level = 1;
  *s = "";
  while ( 1 ) {
    is.get(c);
    if ( !is ) {
      global_error()("closing \'}\' is missing");
      return ERROR;
    }
    if ( c == '{' ) ++level;
    if ( c == '}' ) --level;
    if ( level == 0 ) break;
    (*s) += c;
  }
  return NEXT;
}
// this structure will be used to make an interactive command.
// It reads the input stream, fills vector<vector<vector<string> > > names
// (names[ilayer][ipart][ichamber])
// and then calls HPT_Setup::create_chambers (see below)
struct Get_HPT_Chamber_Names {
  HPT_Setup* setup;
  Get_HPT_Chamber_Names(HPT_Setup* fSetup) : setup(fSetup) {}
  void operator()(istream* is) {
    vector<vector<vector<string> > > names;
    string s1;
    while ( 1 ) {
      Result r = get_block(*is, &s1);
      if ( r == END ) break;
      else if ( r == ERROR ) {
	global_error() << "wrong format in the block of High-Pt chamber names";
	return;
      }      
      names.push_back(vector<vector<string> >());
      istrstream is1(s1.c_str());
      string s2;
      while ( 1 ) {
	Result r1 = get_block(is1, &s2);
	if ( r1 == END ) break;
	else if ( r1 == ERROR ) {
	  global_error()<< "wrong format in the block of High-Pt chamber names";
	  return;
	}      
	names.back().push_back(vector<string>());
	istrstream is2(s2.c_str());
	string name;
	while ( is2 >> name ) names.back().back().push_back(name);
      }
    }
    /*
    for (unsigned int il=0; il<names.size(); ++il) {
      cout << "il=" << il << '\n';
      for (unsigned int ip=0; ip<names[il].size(); ++ip) {
	cout << "ip=" << ip << '\n';
	for (unsigned int ich=0; ich<names[il][ip].size(); ++ich)
	  cout << "ich=" << ich << " " << names[il][ip][ich] << " ";
	cout << '\n';
      }
      cout << '\n';
    }
    */
    setup->create_chambers(names);
  }
};
// This function creates chambers distributed among layers and parts
// according to the given vector<vector<vector<string> > > names
void HPT_Setup::create_chambers(const vector<vector<vector<string> > >& names) {
  mIChambers.resize(names.size());
  for (unsigned int il=0; il<names.size(); ++il) {
    mIChambers[il].resize(names[il].size());
    for (unsigned int ip=0; ip<names[il].size(); ++ip) {
      mIChambers[il][ip].resize(names[il][ip].size());
      for (unsigned int ich=0; ich<names[il][ip].size(); ++ich) {
	pair<iterator, bool> p = insert(value_type(names[il][ip][ich],
						   HPT_Chamber()));
	if ( p.second == false ) {
	  global_error()("chamber name " + names[il][ip][ich] +
			 " is given twice");
	  return;
	}
	mIChambers[il][ip][ich] = p.first;
	mID[names[il][ip][ich]] = HPT_ChamberID(il, ip, ich);
      }
    }
  }
}
	  
// This command uses SetupMenu<HPT_Setup> with only one difference:
// the /create command in this menu is substituted by the command described
// above: it creates not just a collection of chambers without any special 
// ordering, but the chambers which are put into the superlayers and into
// "parts" : inner and outer parts.
void HPT_Setup::read(istream& is) {
  SetupMenu<HPT_Setup> menu(&hpt_setup);
  vector<vector<vector<string> > > names;
  menu("create") = make_interactive_command(Get_HPT_Chamber_Names(this));
  menu(&is);
  if ( global_error() ) return;
  
  //  cross_reference();
}


/* the following function 
     HPT_Setup::cross_reference()
   finds layers/parts automatically using z oordinate of the chambers. Thus it 
   takes just the collection of chambers (without layers/parts information)
   and structures them into layers/parts. It works
   correctly in the design case, when all chambers are installed, outer part is
   shifted from inner downstream in all layers. However it is not used in order
   to give more flexibility in assigning chambers to layers/parts. Currently
   this can be done "by hand" in hpt_geo.dat file using the /create command
   described above.
   */
// LessDouble_Fine is just LessDouble with epsilon = 1.0
// ("Fine" is in comparison with epsilon=20. which will be used later)
struct LessDouble_Fine {
  bool operator()(double x1, double x2) const { return x1 < x2 - 1.; }
};
void HPT_Setup::cross_reference() {
  double epsilon_coarse = 20.; // to distinguish different layers
  // double epsilon_fine = 1.;  
  // which used in LessDouble_Fine above is to distinguish inner/outer inside
  //  layer

  typedef map<double, vector<HPT_Setup::IChamber>, LessDouble_Fine> Zmap1;
  typedef map<double, Zmap1, LessDouble> Zmap;
  LessDouble ld_coarse(epsilon_coarse);
  Zmap zmap(ld_coarse);

  mIChambers.clear();
  mID.clear();
  for (HPT_Setup::IChamber ich = begin(); ich != end(); ++ich) {
    double z = (*ich).second.z;
    zmap[z][z].push_back(ich); // [coarse_group][fine_group]
  }
  
  mIChambers.resize(zmap.size());
  unsigned int il = 0;
  for ( Zmap::iterator iter=zmap.begin(); iter!=zmap.end(); ++iter, ++il ) {
    Zmap1& zmap1 = (*iter).second;
    mIChambers[il].resize(zmap1.size());
    unsigned int ipart = 0;
    for ( Zmap1::iterator iter1=zmap1.begin(); iter1!=zmap1.end();
	  ++iter1, ++ipart ) {
      vector<HPT_Setup::IChamber>& v = (*iter1).second;
      mIChambers[il][ipart] = v;
      for (unsigned int i_ch=0; i_ch<v.size(); ++i_ch) {
	HPT_Setup::IChamber ich = v[i_ch];
	mID[(*ich).first] = HPT_ChamberID(il, ipart, i_ch);
      }
    }
  }
}
  
void HPT_Setup::clear() {
  Setup<HPT_CellT, HPT_SectionT, Dummy, HPT_SetupT>::clear();
  mIChambers.clear();
  mID.clear();
}
