/* This file defines InteractiveMenu and InteractiveCommand classes.

   Both of them are derived from InteractiveItem class.
   InteractiveMenu is a
   map<string, PtrWrapper<InteractiveItem>, less<string> >
   so it can hold other InteractiveMenu's or InteractiveCommand's.
   The whole structure looks like the tree structure of the files
   (InteractiveCommand) in the directories (InteractiveMenu).


   There is a general virtual function 
   void InteractiveItem::operator()(list<string>& args, istream* is);
   1) In InteractiveMenu this function takes the front name from the list
   "args" and finds the corresponding InteractiveItem, so that you go
   to the next submenu (subdirectory) or find the particular 
   InteractiveCommand (file). If the "name" is not found the program tries
   to find the name in the map starting from the same letters as in "name".
   If this can be done unambiguously - the next InteractiveItem is considered
   to be found. After that front "name" is stripped from the list "args"
   and InteractiveItem::operator()(list<string>& args, istream* is) is called
   for the next InteractiveItem (recursively).
   
   2) In InteractiveCommand this function checks that nothing left in 
   the list "args" (otherwise it writes to the "global_error()" object a
   message and returns) and then calls virtual function
   void InteractiveCommand::operator()(istream* is);

   So to use this mechanism you should supply the class derived from
   InteractiveCommand which should define operator()(istream* is) function.

   Example:
   InteractiveMenu menu;
   menu["menu_section1"]["menu_section2"]("your_command") =
     new Your_Class(... any arguments for your class ...);

   Then you should define some parser to strip the information coming 
   from istream to the "menu commands" and "command arguments".
   "Menu commands" should be represented as list<string>& args;
   "command arguments" should be in the (original or newly created)
   istream is.
   Then call
   menu(args, is);
   and that's it. When the program finds the list with the contents:
   {"menu_section1", "menu_section2", "your_command"}
   it invokes Your_Class function object with "is" as a parameter
   (Your_Class::operator()(is)) so that you can e.g. extract the neccessary
   input from "is" and fulfill some appropriate action. 

   To simplify writing the code there is a function
   template <class F>
   make_interactive_command(F f);
   Here F f is any function object which can operate under istream* "is":
   expression f(is) is valid. The make_interactive_command(f) function
   returns the pointer to the new'ly created object (of type
   MakeInteractiveCommand<F>). This type is derived from InteractiveCommand
   and contains 
   virtual void operator()(istream* is) { f(is); }
   virtual InteractiveItem* clone() const; - this function is needed for
   PtrWrapper to act with pointers like with objects (e.g. copy them,
   see ptr_wrap.hh).

   Therefore you can write 
   menu["menu_section1"]["menu_section2"]("your_command") =
     make_interactive_command(f);
   where f should just have the ability to work with istream* via expression
   f(is).
   
   You may also use call_interactively.hh file which defines 
   call_interactively(...) functions. They allow to create function objects
   operating under the istream*'s from any function objects with any arguments.
   The idea is to extracted the arguments (of any types which have the 
   corresponding operator>>(istream&, This_Type)) from the istream and then
   call the corresponding (arbitrary) function object with these arguments. 

   Another tool is in the file istream_menu.hh.
   It defines IstreamMenu class which is derived from InteractiveMenu
   and additionally contains PtrWrapper<InteractiveParser>, where 
   InteractiveParser (defined in interactive_parser.hh) is a general 
   abstract class which fulfills the task of stripping the istream 
   information to the "menu commands" and  "command arguments".
   (you can find examples of classes derived from InteractiveParser in
   interactive_parser.hh).
   
   For the IstreamMenu class the member function run_one_command(istream*)
   is defined. First it calls parser which prepares the "commands" and
   "arguments". Then if there are no errors, parser output information
   is sent to the InteractiveMenu. This call is finally retransmitted
   to the call to the function object which you supplied as it was explained
   above. To process the whole menu you may use operator()(istream*).

   Example how to use all that:

   int foo(int i, int j) { cout << "this is foo: " << i << ", " << j << endl; }
   ...
   IstreamMenu menu;
   menu.parser = new SlashLine_InteractiveParser;
   menu["general_commands"]["special_commands"]..["very_special_commands"]
     ("call_foo") = call_interactively(foo, (int*)0, (int*)0);
   menu(is);

   when the string
   /general_commands/special_commands/../very_special_commands/call_foo 5 10
   is encountered in the istream "*is", the function foo(i,j) will be called
   with aguments i=5, j=10. Here (int*)0 is used in order to give the compiler
   the information about the types of the arguments (you can extract from
   istream the symbol of type char, and supply it as an argument for "foo",
   char will be casted to int).

   Some function objects are defined in the files:
   member_function.hh, assign_to_data_member.hh, binder.hh.
   E.g. you can include calls to member functions using file member_function.hh:

   struct A {
     int i;
     int j;
     A() : i(5), j(10) {}
     void print(int, int) { cout << "this is A: " << i << " " << j << endl; }
     void print1(int, int) const { cout << "this is also A\n"; }
   };
   menu["a"]["b"]("member_function") = 
     call_interactively(member_function(a, A::print, (int*)0, (int*)0), 
		       (int*)0, (int*)0);
   menu["a"]["b"]("const_member_function") = 
     call_interactively(const_member_function(a, A::print1, (int*)0, (int*)0), 
		       (int*)0, (int*)0);

   You can assign values to data members using assign_to_data_member.hh.
   You can bind some arguments in function object using binder.hh
   (e.g. if you have some function object 
   f(a1, a2, a3, a4) receiving 4 arguments, 
   you can construct another function object g(a1, a3, a4) 
   which when invoked calls f(a1, a2=some_const, a3, a4) - a2 is always binded
   to some_const).
*/

// Author V.Balagura (balagura@mail.desy.de)   (10.04.98).

#ifndef _interactive_hh
#define _interactive_hh 1

#include <string>
#include <map>
#include <list>
#include "ptset/ptr_wrap.hh"
#include "ptset/global_error.hh"
#include "ptset/binder.hh"

class InteractiveItemBase {
 protected:
  bool is_menu_;
  InteractiveItemBase() : is_menu_(false) {}
 public:
  bool is_menu() { return is_menu_; }
  virtual ~InteractiveItemBase() {}
};
class InteractiveItem : public InteractiveItemBase {
 public:
  virtual void operator()(list<string>& args, istream* is) = 0;
  virtual InteractiveItem* clone() const = 0;
};
class InteractiveCommand : public InteractiveItem {
 public:
  virtual void operator()(list<string>& args, istream* is);
  virtual void operator()(istream* is) = 0;
};

class InteractiveMenu 
: public InteractiveItem, 
  public map<string, PtrWrapper<InteractiveItem>, less<string> > {
  public:
  InteractiveMenu() { is_menu_ = true; }
  virtual InteractiveItem* clone() const { 
    return new InteractiveMenu(*this); 
  }
  virtual void operator()(list<string>& args, istream* is);

  // use it in the form menu("name") = new Smth
  // if element with "name" is found it is deleted and substituted by new Smth
  // otherwise new Smth is inserted to the map
  PtrWrapper<InteractiveItem>& operator()(const string& s);

  // use it in the form menu["name1"]["name2"]("command_name") = new Smth
  // operator["name"] assumes the object marked as "name" is of 
  // InteractiveMenu type (if not this will produce run time error)
  InteractiveMenu& operator[](const string& s);
};

/* ----- makes InteractiveCommand from function object ----- */
template <class F>
class MakeInteractiveCommand : public InteractiveCommand {
 public:
  F function;

  MakeInteractiveCommand(F f) : function(f) {}
  virtual void operator()(istream* is) { function(is); }
  virtual InteractiveItem* clone() const { 
    return new MakeInteractiveCommand(*this);
  }
  MakeInteractiveCommand<F>* insert_to(InteractiveMenu& menu, 
				       const string& command_name) {
    menu(command_name) = this;
    return this;
  }				       
};
template <class F>
inline MakeInteractiveCommand<F>* make_interactive_command(F f) { 
  return new MakeInteractiveCommand<F>(f); 
}

#endif
