/* Taken from 
  http://www.xraylith.wisc.edu/~khan/software/stl/STL.newbie.html
   - Mumit's STL Newbie guide;
   slightly modified by V.Balagura (balagura@mail.desy.de)   (09.04.98).

   PtrWrapper<T> class implements "clever pointer".
   Usage:
     PtrWrapper<T> ptr1, ptr2;
     ptr1 = new T; // will be deleted automatically
     ptr2 = ptr1; // copy of *(ptr1()) is created with the help of clone()
     PtrWrapper<T> ptr3(ptr1); // copy is created by clone() as well
     PtrWrapper<T> ptr4(new T); // ptr4 holds pointer to new object, it
     // will be deleted automatically by ~PtrWrapper<T>
     T* pt = ptr1(); // gives  normal pointer

   PtrWrapper<T> can be used to construct containers of "clever
   pointers". This in turn gives the possibility to make heterogeneous
   containers. Elements are included via pointers to the common base
   class, e.g.:
     vector<PtrWrapper<BaseClass> > v;
   Since PtrWrapper<T> objects can be copied, assigned etc. like
   the normal objects, the containers made of PtrWrapper<T> can also
   be copied, assigned, deleted etc.. The programmer should not worry
   about deletion or creation of dynamical objects inside the containers
   since all is done automatically inside PtrWrapper<T>. 
   
   Another place to use PtrWrapper<T>: when you'd like to have a memeber
   in a class which "can have various types". E.g. "some figure" which
   can be triangle, rectangle, circle etc.. In this case one may
   define such a member as
     PtrWrapper<Figure> figure;
   where all triangles, rectangles etc. will be derived from the 
   base class Figure. The advantage here is again the fact that you 
   may freely copy, assign etc. the objects of your class with the
   general figure inside.
   

   The class T must support the clone() method.
   --------------------------------------------
   This is a necessary payment if you'd like that PtrWrapper<T>'s
   can be assigned to each other (or constructed from each other).
  
   If you use PtrWrapper to wrap class derived from T 
   (clone method is virtual and creates new instance of derived class)
   class T should have virtual destructor (even empty) so that deleting 
   wrapper will delete the whole derived class object
   */

#ifndef _ptr_wrap_hh
#define _ptr_wrap_hh 1

template <class T>
class PtrWrapper {
public:
  PtrWrapper(T* t_ptr = 0) : t_ptr_(t_ptr) { }
  PtrWrapper(const PtrWrapper<T>& w) {
    t_ptr_ = w() ? w()->clone() : 0;
  }
  ~PtrWrapper() { delete t_ptr_; }
  const T* operator()()  const { return t_ptr_; }
  T* operator()() { return t_ptr_; }
  PtrWrapper<T>& operator= (const PtrWrapper<T>& w) {
    if  (w() != t_ptr_) { // condition is added by V.Balagura
      delete t_ptr_;
      t_ptr_ = w()->clone();
    }
    return *this;
  }
  PtrWrapper<T>& operator= (T* t_ptr) {  // added by V.Balagura
    if  (t_ptr != t_ptr_) {
      delete t_ptr_;
      t_ptr_ = t_ptr;
    }
    return *this;
  }
private:
  T* t_ptr_;
};

template <class T>
bool operator== (const PtrWrapper<T>& w1, const PtrWrapper<T>& w2) {
  return false;
}

template <class T>
bool operator< (const PtrWrapper<T>& w1, const PtrWrapper<T>& w2) {
  return false;
}

#endif

