
// ****************************************************************************
//
//  Routine:      rc99AStandAlone
//  Author:       T.Oest/D.Samtleben
//
//  Description:  stand alone reconstruction
//
//                reco level 0: no reconstruction
//                           1: hitb only
//                           2: reconstruction only
//                           3: hitb + reconstruction
//
//  Modification:  
//  drop RICH hits
//  recoTARGET,... variables no longer static                       27/05/99 TO
//  VDS==SVD for kumac-steering error-message if given subdetector
//  is not known                                                    10/06/99 DS
//  if at least one reconstruction package of a subdetector
//  is called the Rxxx-tables are cleared                           01/07/99 DS
//  bug fix: proper termination of input[]
//  include CATS: recoVDSmodule
//  correct calling sequenz for rc99ATARGET                         26/07/99 TO
//  global alignment transformation added                           06/09/99 TO
//  RICH hits are no longer removed since they are needed for
//  the particle ID                                                 14/09/99 TO 
//  move VDS global alignment transformation to rc99AVDS            20/09/99 TO
//  filling of GTAR bank moved to acReadTARGET                      24/09/99 TO
//  include global VDS alignment transformation                     29/10/99 TO
//  ITR added                                                       07/02/00 TO
//  use ARTE_mark()                                                 12/05/00 AG
//  introduce timing                                                17/05/00 AG
//  timing info into globals                                        18/05/00 AG
//  creation of chrono jobs only once per job                       24/05/00 DS
//  initialize timers to 0                                           7/06/00 AG
//  use more ARTE_mark()                                             8/06/00 AG
//  polishing                                                       16/06/00 AG
//
// ****************************************************************************

// standard includes
#include <iostream.h>
#include <stdlib.h>
#include <stdio.h>

// ARTE includes
#include "arte/arte.hh"
#include "arte/Chrono.hh"
#include "arte/ChronoManager.hh"
#include "farm/farm.hh"
#include "reco/rc99Areco.hh"
#include "reco/rcflags.hh"

// externals
extern int FATIME;

// globals
float time_rc99AVDS  = 0;
float time_rc99AITR  = 0;
float time_rc99AHPT  = 0;
float time_rc99AOTR  = 0;
float time_rc99ARICH = 0;
float time_rc99ATRD  = 0;
float time_rc99AECAL = 0;
float time_rc99AMUON = 0;

// prototypes
void acRSEGinGlobalCoordinates(const int CMPnumber);

//********************************************************************
// rc99AStandAlone
//********************************************************************
int
rc99AStandAlone()
{
  // mark
  ARTE_mark( "rc99AStandAlone 0", 0 );
  
  //===================================================================
  //     remove the Rxxx tables (for real data)
  //===================================================================
  
  //------  remove Rxxx-tables
  if(recoVDS||recoECAL||recoRICH||recoHPT||recoTRD||recoMUON||
     recoOTR){
    ArteTable<RREL>::clear();
    ArteTable<RPNT>::clear();
    ArteTable<RVER>::clear();
    ArteTable<RHIT>::clear();
    ArteTable<RSEG>::clear();
    ArteTable<RTRA>::clear();
    ArteTable<RCCL>::clear();
  }

  //===================================================================
  //     initializations
  //===================================================================
  
  // timers
  static Chrono *cron_rc99AVDS  = 0;
  static Chrono *cron_rc99AITR  = 0;
  static Chrono *cron_rc99AHPT  = 0;
  static Chrono *cron_rc99AOTR  = 0;
  static Chrono *cron_rc99ARICH = 0;
  static Chrono *cron_rc99ATRD  = 0;
  static Chrono *cron_rc99AECAL = 0;
  static Chrono *cron_rc99AMUON = 0;

  static bool first = 1;
  if( first ){
    first = 0;
    ChronoManager CM;
    CM.instance();
    
    cron_rc99AVDS  = CM.createChrono( "rc99AVDS"  );
    cron_rc99AITR  = CM.createChrono( "rc99AITR"  );
    cron_rc99AHPT  = CM.createChrono( "rc99AHPT"  );
    cron_rc99AOTR  = CM.createChrono( "rc99AOTR"  );
    cron_rc99ARICH = CM.createChrono( "rc99ARICH" );
    cron_rc99ATRD  = CM.createChrono( "rc99ATRD"  );
    cron_rc99AECAL = CM.createChrono( "rc99AECAL" );
    cron_rc99AMUON = CM.createChrono( "rc99AMUON" );
  }

  // initialization
  int VDSflag  = 0;
  int ITRflag  = 0;
  int HPTflag  = 0;
  int OTRflag  = 0;
  int RICHflag = 0;
  int TRDflag  = 0;
  int ECALflag = 0;
  int MUONflag = 0;

  //===================================================================
  //     stand-alone modules
  //===================================================================

  // VDS
  if (recoVDS){
    ARTE_mark( "rc99AVDS", 0 );
    cron_rc99AVDS->start();
    VDSflag = rc99AVDS(recoVDS);
    cron_rc99AVDS->stop();
    time_rc99AVDS = cron_rc99AVDS->lastTime();
  }

  // ITR
  if (recoITR){
    ARTE_mark( "rc99AITR", 0 );
    cron_rc99AITR->start();
    ITRflag = rc99AITR(recoITR);
    cron_rc99AITR->stop();
    time_rc99AITR  = cron_rc99AITR->lastTime();
  }

  // HPT
  if (recoHPT){
    ARTE_mark( "rc99AHPT", 0 );
    cron_rc99AHPT->start();
    HPTflag = rc99AHPT(recoHPT);
    cron_rc99AHPT->stop();
    time_rc99AHPT = cron_rc99AHPT->lastTime();
  }

  // OTR
  if (recoOTR){
    ARTE_mark( "rc99AOTR", 0 );
    cron_rc99AOTR->start();
    OTRflag = rc99AOTR(recoOTR);
    cron_rc99AOTR->stop();
    time_rc99AOTR = cron_rc99AOTR->lastTime();
  }

  // RICH
  if (recoRICH){
    ARTE_mark( "rc99ARICH", 0 );
    cron_rc99ARICH->start();
    RICHflag = rc99ARICH(recoRICH);
    cron_rc99ARICH->stop();
    time_rc99ARICH = cron_rc99ARICH->lastTime();
  }

  // TRD
  if (recoTRD){
    ARTE_mark( "rc99ATRD", 0 );
    cron_rc99ATRD->start();
    TRDflag = rc99ATRD(recoTRD);
    cron_rc99ATRD->stop();
    time_rc99ATRD = cron_rc99ATRD->lastTime();
  }

  // MUON
  if (recoMUON){
    ARTE_mark( "rc99AMUON", 0 );
    cron_rc99AMUON->start();
    MUONflag = rc99AMUON(recoMUON);
    cron_rc99AMUON->stop();
    time_rc99AMUON = cron_rc99AMUON->lastTime();
  }

  //===================================================================
  //     transformation of segments/RHIT to global coordinates
  //===================================================================  
  ARTE_mark( "acRSEGinGlobalCoordinates", 0 );
  if (recoVDSmodule & 1) acRSEGinGlobalCoordinates(1);
  if (recoVDSmodule & 2) acRSEGinGlobalCoordinates(65537);
  if (recoRICH)          acRSEGinGlobalCoordinates(67108864);
  
  //===================================================================
  //     ECAL needs tracks for the 2. reconstruction step
  //===================================================================
  if (recoECAL){
    ARTE_mark( "rc99AECAL", 0 );
    cron_rc99AECAL->start();
    ECALflag = rc99AECAL(recoECAL);
    cron_rc99AECAL->stop();
    time_rc99AECAL = cron_rc99AECAL->lastTime();
  }

  //===================================================================
  //     print
  //===================================================================
  if( FATIME ){

    float sum =
      time_rc99AVDS  +
      time_rc99AITR  +
      time_rc99AHPT  +
      time_rc99AOTR  +
      time_rc99ARICH +
      time_rc99ATRD  +
      time_rc99AECAL +
      time_rc99AMUON;

    cout << " RECO: "
	 << "vds="  << time_rc99AVDS  << " "
	 << "itr="  << time_rc99AITR  << " "
	 << "hpt="  << time_rc99AHPT  << " "
	 << "otr="  << time_rc99AOTR  << " "
	 << "rich=" << time_rc99ARICH << " "
	 << "trd="  << time_rc99ATRD  << " "
	 << "ecal=" << time_rc99AECAL << " "
	 << "muon=" << time_rc99AMUON << " "
	 << "sum="  << sum            << endl;
  }
  
  // mark
  ARTE_mark( "rc99AStandAlone 1", 0 );
  
  return 0;  
}
